/*
 * Copyright (C) 2008 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.tools.layoutlib.create;

import java.util.HashMap;

/**
 * Allows stub methods from LayoutLib to be overriden at runtime.
 */
public final class OverrideMethod {

    /**
     * Interface to allow a method invocation to be listend upon.
     */
    public interface MethodListener {
        /**
         * A stub method is being invoked.
         * <p/>
         * Known limitation: caller arguments are not available. Return value cannot be set.
         *  
         * @param signature The signature of the method being invoked, composed of the
         *                  binary class name followed by the method descriptor (aka argument
         *                  types). Example: "com/foo/MyClass/InnerClass/printInt(I)V".
         * @param isNative True if the method was a native method.
         * @param caller The calling object. Null for static methods, "this" for instance methods.
         */
        public void onInvoke(String signature, boolean isNative, Object caller);
    }

    /** Map of method overridden. */
    private static HashMap<String, MethodListener> sMethods = new HashMap<String, MethodListener>();
    /** Default listener for all method not listed in sMethods. Nothing if null. */
    private static MethodListener sDefaultListener = null;
    
    /**
     * Sets the default listener for all methods not specifically handled.
     * Null means to do nothing.
     */
    public static void setDefaultListener(MethodListener listener) {
        sDefaultListener = listener;
    }

    /**
     * Defines or reset a listener for the given method signature.
     * 
     * @param signature The signature of the method being invoked, composed of the
     *                  binary class name followed by the method descriptor (aka argument
     *                  types). Example: "com/foo/MyClass/InnerClass/printInt(I)V"
     * @param listener The new listener. Removes it if null.
     */
    public static void setMethodListener(String signature, MethodListener listener) {
        if (listener == null) {
            sMethods.remove(signature);
        } else {
            sMethods.put(signature, listener);
        }
    }
    
    /**
     * Invoke the specific listener for the given signature or the default one if defined. 
     * <p/>
     * Note: this is not intended to be used by the LayoutLib Bridge. It is intended to be called
     * by the stubbed methods generated by the LayoutLib_create tool.
     * 
     * @param signature The signature of the method being invoked, composed of the
     *                  binary class name followed by the method descriptor (aka argument
     *                  types). Example: "com/foo/MyClass/InnerClass/printInt(I)V".
     * @param isNative True if the method was a native method.
     * @param caller The calling object. Null for static methods, "this" for instance methods.
     */
    public static void invoke(String signature, boolean isNative, Object caller) {
        MethodListener i = sMethods.get(signature);
        if (i != null) {
            i.onInvoke(signature, isNative, caller);
        } else if (sDefaultListener != null) {
            sDefaultListener.onInvoke(signature, isNative, caller);
        }
    }
}
